import logging
import os.path
import subprocess
import threading
import time

import dbus
import dbus.service

import xdg.BaseDirectory
from xdg.DesktopEntry import DesktopEntry

from .notification import Notification
from .rofi_gui import Operation

DBUS_PATH = "org.freedesktop.Notifications"

logger = logging.getLogger(__name__)


class Switcher(threading.Thread):
    def run(self):
        subprocess.call(["switch-to-urgent.py"])


def xdg_name_and_icon(app):
    entry = DesktopEntry()
    for directory in xdg.BaseDirectory.xdg_data_dirs:
        path = os.path.join(directory, "applications", f"{app}.desktop")
        if os.path.exists(path):
            entry.parse(path)
            return entry.getName(), entry.getIcon()
    return None, None


class NotificationFetcher(dbus.service.Object):
    _id = 1

    def __init__(self, dm, gui):
        self.dm = dm
        self.gui = gui
        self.context = []

        if len(self.dm.tree):
            self._id = self.dm.tree.last().id + 1

        name = dbus.service.BusName(DBUS_PATH, dbus.SessionBus())
        super().__init__(name, "/org/freedesktop/Notifications")

    def _update_context(self):
        new_context = []
        p = self.dm.tree
        for key in self.context:
            if key not in p.notifications:
                break
            new_context.append(key)
            p = p.notifications[key]
        if self.context == new_context:
            logger.info("Context did not change")
            return False

        logger.info("Context updated.")
        self.context = new_context
        return True

    def _show_notifications(self, row=0):
        notifications = self.dm.get_context(self.context).notifications

        items = sorted(
            [
                (k, v) if len(v) > 1 else (v.last().id, v.last())
                for k, v in notifications.items()
            ],
            key=lambda x: (-x[1].urgency, -x[1].last().created_at),
        )

        selected, op = self.gui.show_notifications([item[1] for item in items], row)
        logger.info(
            f"Selection is {None if selected is None else items[selected][0]}"
            f", operation is {op}."
        )

        if op == Operation.EXIT:
            if self.context:
                self.context.pop()
                self._show_notifications()
            return

        if selected is None:
            logger.info(f"DEBUG THIS {items}")
            return

        key, notification = items[selected]

        if op == Operation.SELECT:
            if isinstance(notification, Notification):
                logger.info("Selection is a singleton. Invoking default action.")
                self.context = self.dm.map[notification.id]
                Switcher().start()
                time.sleep(0.25)
                self.ActionInvoked(notification.id, "default")
            else:
                logger.info("Selection is a cluster. Expanding.")
                self.context.append(key)
                self._show_notifications()
        elif op == Operation.DELETE:
            self.dm.remove_notification(key, self.context)

            is_new_context = self._update_context()
            row = 0 if is_new_context or len(notifications) == 1 else selected

            self._notifications_updated()

            if len(self.dm.tree):
                self._show_notifications(row)

    @dbus.service.method(DBUS_PATH, in_signature="susssasa{ss}i", out_signature="u")
    def Notify(
        self,
        app_name,
        replaces_id,
        app_icon,
        summary,
        body,
        actions,
        hints,
        expire_timeout,
    ):
        logger.info(
            "Received notification :\n"
            f'app_name:"{app_name}" '
            f"replaces_id:{replaces_id} "
            f'app_icon:"{app_icon}" '
            f'summary:"{summary}" '
            f'body:"{body}" '
            f'actions:"{actions}" '
            f'hints:"{hints}" '
            f"expire_timeout:{expire_timeout}"
        )

        if replaces_id > 0:
            id = replaces_id
        else:
            id = self._id
            self._id += 1

        app = icon = None
        if "desktop-entry" in hints and not (app_name and app_icon):
            app, icon = xdg_name_and_icon(hints["desktop-entry"])
            app_name = app_name or app or hints["desktop-entry"].split(".")[-1]

        if not app_icon or app_icon.startswith("file://"):
            if icon:
                app_icon = icon
            elif "image-path" in hints:
                app_icon = hints["image-path"]

        notification = Notification(
            id=id,
            app_name=app_name,
            app_icon=app_icon,
            summary=summary,
            body=body,
            actions=actions,
            created_at=time.time_ns(),
        )

        if expire_timeout > 0:
            notification.expire_at = notification.created_at + expire_timeout

        if "urgency" in hints:
            notification.urgency = (
                int(hints["urgency"])
                if isinstance(hints["urgency"], str)
                else hints["urgency"].real
            )

        self.dm.add_notification(notification)
        self._notifications_updated()
        return id

    @dbus.service.method(DBUS_PATH, in_signature="", out_signature="as")
    def GetCapabilities(self):
        return [
            "actions",
            "body",
            "body-hyperlinks",
            "body-images",
            "body-markup",
            "icon-static",
        ]

    def _notifications_updated(self):
        self.NotificationsUpdated(len(self.dm.tree), self.dm.tree.urgency or 0)

    @dbus.service.method(DBUS_PATH, in_signature="u", out_signature="")
    def CloseNotification(self, id):
        notification = self.dm.get_context_by_id(id).notifications[id]
        logger.info(f"Received CloseNotification request for {notification}")

        if notification.closeable():
            self.dm.remove_notification(id)
            self._update_context()
            self.NotificationClosed(id, 3)
            self._notifications_updated()
        else:
            logger.info(
                " but notification is configured as non-closeable, "
                "so it won't be automatically deleted from the view."
            )

    @dbus.service.method(DBUS_PATH, in_signature="", out_signature="ssss")
    def GetServerInformation(self):
        return "notification_fetcher", "github.com/sencer", "0.0.0", "1"

    @dbus.service.method(DBUS_PATH, in_signature="", out_signature="")
    def ShowNotifications(self):
        self.context = []
        self._show_notifications()

    @dbus.service.method(DBUS_PATH, in_signature="", out_signature="uu")
    def ShowNotificationCount(self):
        return len(self.dm.tree), self.dm.tree.urgency or 0

    @dbus.service.method(DBUS_PATH, in_signature="", out_signature="")
    def SignalNotificationCount(self):
        self._notifications_updated()

    @dbus.service.method(DBUS_PATH, in_signature="", out_signature="s")
    def DumpNotifications(self):
        self.dm.dump()
        return str(self.dm.tree)

    @dbus.service.signal(DBUS_PATH, signature="uu")
    def NotificationClosed(self, id, reason):
        logger.info(f"NotificationClosed signalled for {id} due to {reason}.")

    @dbus.service.signal(DBUS_PATH, signature="us")
    def ActionInvoked(self, id, action):
        logger.info(f"ActionInvoked with action {action} signalled for {id}.")

    @dbus.service.signal(DBUS_PATH, signature="uu")
    def NotificationsUpdated(self, num, urgency):
        logger.info(f"Notifications updated.")



import pickle
import threading

from .notification import Notification, NotificationCluster


class DataManager(threading.Thread):
    def __init__(self, configs, dump_path):
        super().__init__()

        self.tree = NotificationCluster()
        self.map = dict()

        self.lock = threading.Lock()
        self.configs = configs
        self.dump_path = dump_path

        try:
            for notification in pickle.load(open(dump_path, "rb")):
                self.add_notification(notification)
        except:
            pass

    def _recursive_add_notification(cluster, notification, keys, i=0):
        if i == len(keys):
            return

        if keys[i] not in cluster.notifications:
            cluster.notifications[keys[i]] = NotificationCluster()

        DataManager._recursive_add_notification(
            cluster.notifications[keys[i]], notification, keys, i + 1
        )
        cluster.add(keys[i], notification)

    def add_notification(self, notification):
        for config in self.configs:
            if config.should_apply(notification):
                config.update_notification(notification)
                notification.config = config
                break

        keys = notification.keys()

        if notification.id in self.map:
            self.remove_notification(notification.id)

        with self.lock:
            self.map[notification.id] = keys
            DataManager._recursive_add_notification(
                self.tree, notification, [*keys, notification.id]
            )

    def _recursive_remove_notification(cluster, keys, i=0):
        key = keys[i]
        last_key = i == len(keys) - 1
        has_key = key in cluster.notifications
        if last_key and not has_key:
            # Short-cutted view, descend
            key = list(cluster.notifications.keys())[0]
            i -= 1

        stop_case = last_key and has_key
        if stop_case:
            cluster_to_delete = cluster.notifications[key]
            last = cluster_to_delete.last()
            urgency = cluster.urgency
            nremoved = len(cluster_to_delete)
        else:
            nremoved, last, urgency = DataManager._recursive_remove_notification(
                cluster.notifications[key], keys, i + 1
            )

        if len(cluster.notifications[key]) == 0 or stop_case:
            del cluster.notifications[key]

        cluster._len -= nremoved

        if last is cluster._last:
            cluster._last = None

        if urgency == cluster._urgency:
            cluster._urgency = None

        return nremoved, last, urgency

    def remove_notification(self, id, context=()):
        with self.lock:
            if isinstance(id, int):
                context = self.map.pop(id)
            else:
                for leaf in self.get_context(context).notifications[id].leafs():
                    self.map.pop(leaf.id)

            DataManager._recursive_remove_notification(self.tree, [*context, id], i=0)

    def get_context_by_id(self, id):
        return self.get_context(self.map[id])

    def get_context(self, context=()):
        p = self.tree
        for key in context:
            p = p.notifications[key]

        while len(p.notifications) == 1:
            child = list(p.notifications.values())[0]

            if isinstance(child, Notification):
                break

            p = child

        return p

    def dump(self):
        pickle.dump(
            [notification.strip() for notification in self.tree.leafs()],
            open(self.dump_path, "wb"),
        )



import unittest

from i3notifier.config import Config
from i3notifier.data_manager import DataManager
from i3notifier.notification import Notification, NotificationCluster


class DummyConfig(Config):
    def get_keys(notification):
        return (notification.app_name, notification.summary)


class TestDataManager(unittest.TestCase):
    def __init__(self, *args, **kwargs):
        super(TestDataManager, self).__init__(*args, **kwargs)
        self.notifications = [
            Notification(1, "A3", "icon", "1", "113", ["dflt"], 1595608250375722879, 1),
            Notification(2, "A1", "icon", "1", "111", ["dflt"], 1595608250375722880, 1),
            Notification(3, "A1", "icon", "2", "121", ["dflt"], 1595608250375722881, 1),
            Notification(4, "A1", "icon", "1", "112", ["dflt"], 1595608250375722882, 1),
            Notification(5, "A2", "icon", "1", "211", ["dflt"], 1595608250375722884, 1),
            Notification(6, "A2", "icon", "2", "212", ["dflt"], 1595608250375722885, 1),
            Notification(7, "A1", "icon", "1", "113", ["dflt"], 1595608250375722886, 1),
        ]

        self.dm = DataManager([DummyConfig], "/dev/null")
        for notification in self.notifications:
            self.dm.add_notification(notification)

    def test_get_context_by_id(self):
        self.assertIs(
            self.dm.get_context_by_id(7),
            self.dm.tree.notifications["A1"].notifications["1"],
        )

    def test_get_context(self):
        self.assertIs(
            self.dm.get_context(("A1", "1")),
            self.dm.tree.notifications["A1"].notifications["1"],
        )

    def test_get_nocontext(self):
        self.assertIs(self.dm.get_context(), self.dm.tree)

    def test_leafs(self):
        self.assertCountEqual(self.dm.tree.leafs(), self.notifications)
        self.assertCountEqual(
            self.dm.get_context(("A1", "1")).leafs(),
            [self.notifications[1], self.notifications[3], self.notifications[6],],
        )

    def test_add_notification(self):
        self.assertEqual(len(self.dm.tree), 7)
        self.assertEqual(len(self.dm.get_context(("A2",))), 2)
        self.assertIs(self.dm.tree.last(), self.notifications[-1])

        notification = Notification(
            8, "A2", "icon", "1", "212", ["dflt"], 1595608250375722891, 1
        )

        self.dm.add_notification(notification)
        self.assertEqual(len(self.dm.tree), 8)
        self.assertEqual(len(self.dm.tree.notifications["A2"]), 3)
        self.assertIs(self.dm.get_context().last(), notification)
        self.assertIs(self.dm.get_context(("A2",)).last(), notification)

    def test_overwrite_notification(self):
        notification = Notification(
            3, "A1", "icon", "1", "212", ["dflt"], 1595608250375722891, 1
        )

        self.dm.add_notification(notification)
        self.assertEqual(len(self.dm.get_context()), len(self.notifications))
        self.assertEqual(len(self.dm.get_context(("A1",))), 4)
        self.assertIs(self.dm.get_context().last(), notification)
        self.assertIs(self.dm.get_context(("A1",)).last(), notification)

    def test_remove_notification(self):
        self.assertEqual(len(self.dm.get_context(("A1",))), 4)
        self.assertEqual(len(self.dm.tree), len(self.notifications))

        self.assertIs(self.dm.get_context().last(), self.notifications[6])
        self.assertIs(self.dm.get_context(("A1",)).last(), self.notifications[6])

        self.dm.remove_notification("1", ("A1",))

        self.assertEqual(len(self.dm.get_context(("A1",))), 1)
        self.assertEqual(len(self.dm.tree), 4)

        self.assertIs(self.dm.tree.last(), self.notifications[5])
        self.assertIs(self.dm.get_context(("A1",)).last(), self.notifications[2])

    def test_remove_notification_integer(self):
        self.assertEqual(len(self.dm.get_context(("A1",))), 4)
        self.assertEqual(len(self.dm.tree), 7)

        self.assertIs(self.dm.get_context().last(), self.notifications[6])
        self.assertIs(self.dm.get_context(("A1",)).last(), self.notifications[6])

        self.dm.remove_notification(7)

        self.assertEqual(len(self.dm.get_context(("A1",))), 3)
        self.assertEqual(len(self.dm.tree), 6)

        self.assertIs(self.dm.tree.last(), self.notifications[5])
        self.assertIs(self.dm.get_context(("A1",)).last(), self.notifications[3])

    def test_remove_shortcutted(self):
        self.dm.remove_notification(1)
        self.assertCountEqual(self.dm.tree.leafs(), self.notifications[1:])

        self.assertEqual(len(self.dm.tree), 6)
        self.assertNotIn("A3", self.dm.tree.notifications)


if __name__ == "__main__":
    unittest.main()


if __name__ == "__main__":
    unittest.main()



import copy
import itertools
from enum import IntEnum

from .config import Config


class Urgency(IntEnum):

    LOW = 0
    MEDIUM = 1
    CRITICAL = 2


class Notification:
    def __init__(
        self,
        id,
        app_name,
        app_icon,
        summary,
        body,
        actions,
        created_at,
        expires_at=None,
        urgency=0,
    ):
        self.id = id
        self.app_name = app_name
        self.app_icon = app_icon
        self.summary = summary
        self.body = body
        self.actions = actions
        self.created_at = created_at
        self.expires_at = expires_at
        self.urgency = urgency
        self.config = Config

    def formatted(self):
        return self.config.format_notification(self)

    def keys(self):
        return self.config.get_keys(self)

    def closeable(self):
        return self.config.closeable(self)

    def strip(self):
        return Notification(
            self.id,
            self.app_name,
            self.app_icon,
            self.summary,
            self.body,
            self.actions,
            self.created_at,
            self.expires_at,
            self.urgency,
        )

    def __len__(self):
        return 1

    def last(self):
        return self

    def leafs(self):
        return [self]

    def __repr__(self):
        return (
            "<Notification: "
            f"id:{self.id} "
            f'app_name:"{self.app_name}" '
            f'app_icon:"{self.app_icon}" '
            f'summary:"{self.summary}" '
            f'body:"{self.body}" '
            f"actions:{self.actions} "
            f"urgency:{self.urgency} "
            f"created_at:{self.created_at} "
            f"expires_at:{self.expires_at}>"
        )

    def __str__(self):
        return self.__repr__()


class NotificationCluster:
    def __init__(self):
        self.notifications = dict()
        self._last = None
        self._len = 0
        self._urgency = None

    @property
    def urgency(self):

        if self._urgency is None and self.notifications:
            self._urgency = self.last().urgency

        return self._urgency or 0

    def formatted(self):
        if len(self) == 1:
            return self.last().formatted()

        dummy = copy.deepcopy(self.last())
        dummy.app_name = f"{dummy.app_name} ({len(self)})"
        return dummy.formatted()

    def reset(self):
        self._len = 0
        self._urgency = None
        self._last = None

    def add(self, key, notification):

        if self._last is None or notification.urgency >= self.last().urgency:
            self._last = notification
        self._urgency = self.last().urgency
        self._len += 1

        if isinstance(key, int):
            self.notifications[key] = notification

    def remove(self, key):
        if self.urgency == self.notifications[key].urgency:
            self._urgency = None

        if self.notifications[key] == self.last():
            self._last = None

        self._len -= len(self.notifications[key])

        del self.notifications[key]

    def last(self):
        # This is called last for historical reasons, where I represented a cluster
        # by the last notification in it. Now urgency trumps creation time. So this
        # is kind of 'best' rather than 'last'
        if self._last is None and self.notifications:
            self._last = max(
                self.notifications.values(),
                key=lambda x: (x.urgency, x.last().created_at),
            ).last()

        return self._last

    def __len__(self):
        self._len = self._len or sum(len(n) for n in self.notifications.values())
        return self._len or 0

    def leafs(self):
        return list(
            itertools.chain.from_iterable(
                v.leafs() for v in self.notifications.values()
            )
        )

    def __str__(self):
        return str(self.notifications)

    def __repr__(self):
        return repr(self.notifications)



from datetime import datetime

from .utils import strip_tags


class Config:
    def should_apply(notification):
        # To group notifications and apply certain formatting inherit this
        # class and make sure should_apply(notification) is the first Config
        # that returns True to those notifications.
        return True

    def closeable(notification):
        # It is better to ignore the close request for some notifications
        return True

    def get_keys(notification):
        # Returns a tuple, each element of which corresponds to the group
        # name that the notification should be assigned at the
        # corresponding nesting level.
        return (notification.app_name or "_", notification.body)

    def update_notification(notification):
        # Edit the notification before saving it. By default no edits are
        # made.
        return notification

    def format_notification(notification):
        time = datetime.fromtimestamp(notification.created_at // 1e9).strftime("%H:%M")
        # Format the notification
        text = f"<b>{strip_tags(notification.summary)}</b> "
        text += f"<small>{time}</small>"
        text += f"<small> {notification.app_name}</small>"

        if notification.body:
            text += (
                "\n<i>"
                + strip_tags(notification.body.replace(r"\n", "").strip())
                + "</i>"
            )

        if notification.app_icon:
            return (
                text.encode("utf-8")
                + b"\x00icon\x1f"
                + notification.app_icon.encode("utf-8")
            )
        else:
            return text.encode("utf-8")



from distutils.core import setup

setup(
    name="i3-notifier",
    version="0.9",
    description="A notification daemon for i3",
    author="Sencer Selcuk",
    packages=["i3notifier", "tests"],
    scripts=["bin/i3-notifier", "bin/switch-to-urgent.py"],
)



import html
from html.parser import HTMLParser
from io import StringIO


class MLStripper(HTMLParser):
    def __init__(self):
        super().__init__()
        self.reset()
        self.strict = False
        self.convert_charrefs = True
        self.text = StringIO()

    def handle_data(self, d):
        self.text.write(d)

    def get_data(self):
        return self.text.getvalue()


def strip_tags(data):
    s = MLStripper()
    s.feed(html.escape(data))
    return s.get_data()



import subprocess
from enum import Enum


class Operation(Enum):
    SELECT = 0
    EXIT = 1
    DELETE = 10


class RofiGUI:

    _separator = b"\x01"
    _args = [
        "-dmenu",
        "-p",
        "Notifications",
        "-eh",
        "2",
        "-width",
        "-40",
        "-markup-rows",
        "-i",
        "-format",
        "i",
        "-sep",
        r"\x01",
        "--kb-accept-custom",
        "",
        "--kb-accept-alt",
        "",
        "-kb-custom-1",
        "Ctrl+Delete",
        "-show-icons",
        "-columns",
        "1",
        "-lines",
        "10",
    ]

    def __init__(self, *args, cmd=None):
        self.cmd = cmd or "rofi"
        self._args.extend(args)

    def show_notifications(self, notifications, row=0):

        formatted_notifications = [
            notification.formatted() for notification in notifications
        ]

        urgent = [
            str(i)
            for i, notification in enumerate(notifications)
            if notification.urgency == 2
        ]

        proc = subprocess.Popen(
            [self.cmd]
            + self._args
            + ["-selected-row", str(row)]
            + (["-u", ",".join(urgent)] if urgent else []),
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
        )

        proc.stdin.write(self._separator.join(formatted_notifications))
        proc.stdin.close()

        maybe_selection = (lambda x: int(x) if x else None)(
            proc.stdout.read().decode("utf-8")
        )
        operation = Operation(proc.wait())
        return maybe_selection, operation
